import numpy as np
import pandas as pd

from azapy.MkT.MkTcalendar import NYSEgen

class _RiskEngine():
    """
    Base class. Derive class needs to implement\n
        getWeights \n

    Methods:
        * getPositions
        * set_rrate
        * set_mktdata
    """

    def __init__(self, mktdata=None, colname='adjusted',
                 freq='Q', hlength=3.25, calendar=None):
        """
        Constructor

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`, optional;
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function. The default is `None`.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights
            calibration. The default is 'adjusted'.
        `freq` : `str`, optional;
            Rate of returns horizon in number of business day. It could be
            'Q' for quarter or 'M' for month. The default is 'Q'.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A
            fractional number will be rounded to an integer number of months.
            The default is `3.25`.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will 
            be set to NYSE business calendar.
            The default is `None`.

        Returns
        -------
        The object.
        """
        if mktdata is not None:
            self.set_mktdata(mktdata, colname, freq, hlength)

    def set_rrate(self, rrate):
        """
        Sets portfolio components historical rates of returns in the format
        "date", "symbol1", "symbol2", etc.

        Parameters
        ----------
        `rrate` : `pandas.DataFrame`;
            The portfolio components historical rates of returns.
            If it is not `None`, it will overwrite the rrate computed in the
            constructor from mktdata. The default is `None`.
        Returns
        -------
        `None`
        """
        self.nn, self.mm = rrate.shape
        self.muk = rrate.mean()
        self.rrate = rrate

    def set_mktdata(self, mktdata, colname='adjusted',
                    freq='Q', hlength=3.25, calendar=None):
        """
        Sets historical market data. It will overwrite the choice made in the
        constructor.

        Parameters
        ----------
        `mktdata` : `pandas.DataFrame`;
            Historic daily market data for portfolio components in the format
            returned by `azapy.mktData` function.
        `colname` : `str`, optional;
            Name of the price column from mktdata used in the weights
            calibration. The default is `'adjusted'`.
        `freq` : `str`, optional;
            Rate of returns horizon in number of business day. it could be
            'Q' for quarter or 'M' for month. The default is `'Q'`.
        `hlength` : `float`, optional;
            History length in number of years used for calibration. A
            fractional number will be rounded to an integer number of months.
            The default is `3.25` years.
        `calendar` : `numpy.busdaycalendar`, optional;
            Business days calendar. If is it `None` then the calendar will 
            be set to NYSE business calendar.
            The default is `None`.

        Returns
        -------
        `None`
        """
        if mktdata is None:
            return

        self.mktdata = mktdata

        periods = 63 if freq == 'Q' else 21

        if calendar is None:
            calendar = NYSEgen()

        edate = mktdata.index[-1]
        sdate = edate - pd.offsets.DateOffset(months=round(hlength * 12, 0))
        sdate = np.busday_offset(sdate.date(),
                                 0, roll='backward',  busdaycal=calendar)
        rrate = mktdata.pivot(columns='symbol', values=colname)[sdate:]
        self.last_data = rrate.iloc[-1]
        rrate = rrate.pct_change(periods=periods).dropna()

        self.set_rrate(rrate)

    def getPositions(self, nshares=None, cash=0., ww=None):
        """
        Computes the number of shares according to the weights

        Parameters
        ----------
        `nshares` : `pandas.Series`, optional;
            Number of shares per portfolio component. A missing component
            entry will be considered `0`. A `None` value assumes that all
            components entries are `0`. The name of the components must be
            present in the mrkdata. The default is `None`.
        `cash` : `float`, optional;
            Additional cash to be considered in the overall capital. A
            negative entry assumes a reduction in the total capital
            available for rebalance. The remaining capital must be >0.
            The default is `0`.
        `ww` : `pandas.Series`, optional;
            External portfolio weights. If it not set to `None` these
            weights will overwrite the calibrated weights.
            The default is `None`.

        Returns
        -------
        `pandas.DataFrame`;
            The rolling information.

        Columns:

            - `'old_nsh'` :
                the initial number of shares per portfolio
                component as well as additional cash position. These are
                present in the input.
            - `'new_nsh'` :
                the new number of shares per component plus the
                residual cash (due to the rounding to an integer number of
                shares). A negative entry means that the investor needs to
                add more cash to cover for the number of share
                roundup shortfall. It has a small value.
            - `'diff_nsh'` :
                the number of shares that needs to be
                both/sold to rebalance the portfolio positions.
            - `'weights'` :
                portfolio weights used for rebalancing. The 'cash'
                entry is the new portfolio value.
            - `'prices'` :
                the share prices used for rebalances evaluations.

        Note: Since the prices are closing prices, the rebalance can be
        executed next business. Additional cash slippage may occur due
        to share price differential between the previous day closing and
        execution time.

        """
        ns = pd.Series(0, index=self.rrate.columns)
        if nshares is not None:
            ns = ns.add(nshares, fill_value=0)

        if len(self.rrate.columns) != len(ns):
            raise ValueError("nshares must by a subset "
                             f"of {self.rrate.columns}")

        pp = self.last_data[self.rrate.columns.to_list()]

        cap = ns.dot(pp) + cash
        if ww is None:
            ww = self.getWeights()
        else:
            ww0 = pd.Series(0, index=self.rrate.columns)
            ww = ww0.add(ww, fill_value=0)
            if len(self.rrate.columns) != len(ns):
                raise ValueError("ww: wrong component names - "
                                 f"must be among {self.rrate.columns}")

        newns = (ww / pp * cap).round(0)
        newcash = cap - newns.dot(pp)

        ns['cash'] = cash
        newns['cash'] = newcash
        ww['cash'] = cap - newcash
        pp['cash'] = 1.

        res = pd.DataFrame({'old_nsh': ns,
                            'new_nsh': newns,
                            'diff_nsh': newns - ns,
                            'weights' : ww.round(6),
                            'prices': pp})

        return res

    # need to be implemented by the derive classes
    def getWeights(self):
        pass



import numpy as np
import pandas as pd
import cvxopt as cx
import warnings

def gamblingKelly(pp=[0.6]):
    """
    Compute the Kelly allocation for multiple binary games.

    Parameters
    ----------
    `pp` : `list`, optional;
        List of wining probabilities for each game. The default is `[0.6]`.

    Returns
    -------
    `pandas.Series`
        Bet sizes as percentage of the capital=1.
    """
    pp = np.array(pp)
    assert all((pp > 0) & (pp < 1)), "all pp must by in (0,1)"
    nn = len(pp)
    
    # local function 
    def _ffval(w, p):
        n=len(p)
        z = 0.
        df = np.zeros(n)
        dh = np.zeros((n,n))
        for x in range(2**n):
            ppro = 1.
            wsum = 1.
            for k in range(n):
                if x & (1 << k):
                    ppro *= p[k]
                    wsum += w[k]
                else:
                    ppro *= 1 - p[k]
                    wsum -= w[k]
                    
            z += ppro * np.log(wsum)
            for j in range(n):
                jsi = 1 if x & (1 << j) else -1
                term = jsi * ppro / wsum
                df[j] +=  term
                for i in range(j + 1):
                    isi = 1 if x & (1 << i) else -1
                    dh[i,j] -= isi * term / wsum
                    dh[j,i] = dh[i,j]
                    
        return z, df, dh
    
    # local function
    def _F(x=None, z=None):
        if x is None: 
            return 0, cx.matrix(0.5/nn, (nn, 1))
        
        xx = np.array(x.T)[0]
        # ensure that the log can be computed
        if np.sum(xx) >= 1.:
            xx = xx / np.sum(xx) * 0.999999
        f, df, dh = _ffval(xx, pp)
        val = -f
        DF = cx.matrix(-df).T
        if z is None: 
            return val, DF
        
        H = cx.matrix(-z[0] * dh)
        
        return val, DF, H

    # build G and h
    icol = list(range(nn)) * 2
    irow = list(range(nn)) + [nn] * nn
    data = [-1.] * nn + [1.] * nn
    G = cx.spmatrix(data, irow, icol, (nn + 1, nn))
    h = cx.matrix([0.] * nn + [1.])
    dims = {'l': nn + 1, 'q': [], 's': []}
    
    sol = cx.solvers.cp(F=_F, G=G, h=h, dims=dims, 
                        options={'show_progress': False})
    
    if 'optimal' not in sol['status']:
        warnings.warn(f"cannot find a good solution msg: {sol['status']}")
        
    return pd.Series(sol['x']).round(4) * 100



import setuptools
#from azapy import __version__

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setuptools.setup(
    name="azapy",
    version="1.1.1",
    author="Mircea Marinescu",
    author_email="mircea.marinescu@outlook.com",
    description="Financial Portfolio Optimization Algorithms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mircea-MMXXI/azapy.git",
    project_urls={
        "Documentation": "https://azapy.readthedocs.io/en/latest",
        "Source": "https://github.com/Mircea-MMXXI/azapy",
        "Bug Tracker": "https://github.com/Mircea-MMXXI/azapy/issues",
    },
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(),
    python_requires=">=3.8",
    install_requires=[
          'numpy',
          'pandas',
          'scipy',
          'plotly',
          'matplotlib',
          'requests',
          'ecos',
          'pandas_market_calendars',
          'cvxopt',
          'ta',
          'yfinance',
          'statsmodels'
      ],
)



"""
Contains:
    
    - add_cash_security : and a cash like security
    - update_all_MkTData : update mkt data saved in a directory
"""

import pandas as pd
import os
import warnings
from azapy.MkT.MkTreader import MkTreader

def add_cash_security(data, name='_CASH_', value=1.):
    '''
    Add a cash like security to the MkT data.

    Parameters
    ----------
    `data` : `pandas.DataFrame` or `dict`;
        MkT data to add a cash like positions.
    `name` : `str`, optional;
        The symbol of the cash like security. Must be different than any 
        symbol in data. Note that CASH is a valid stock symbol.
        The default is '_CASH_'.
    `value` : `float`, optional;
        Nominal value of the cash like security. It is 
        constant over time (no dividends and no splits).
        The default is `1`.

    Returns
    -------
    The new MkT data updated with the cash like security 
    (in the same format of 'data' input format, i.e. `pandas.DataFrame or 
     'dict').
    '''
    if isinstance(data, pd.core.frame.DataFrame):
        cash = pd.DataFrame(value, index=data.index.unique(), 
                            columns=data.columns)
        cash['symbol'] = name
        cash['divd'] = 0.
        return pd.concat([data, cash])
    
    elif isinstance(data, dict):
        ld = list(data.values())[0]
        cash = pd.DataFrame(value, index=ld.index, columns=ld.columns)
        cash['symbol'] = name
        cash['divd'] = 0.
        data[name] = cash
        return data
    
    else:
        raise TypeError(f"data: wrong type {type(data)} "
                        + "must be pandas.DataFrame or dict")
        
    
def update_all_MkTData(mktdir, source=None, api_key=None, param=None,
                       except_file=[], verbose=True):
    '''
    Updates all mkt data saved in a directory.

    Parameters
    ----------
    `mktdir` : `str`;
        Mkt data directory.
    `source` : `str`, optional;
        Mkt data provider. 
        For more details see the `azapy.MkTreader.get` function doc. 
        The default is `None`.
    `api_key` : `str`, optional;
        Mkt data provider API key.
        For more details see the `azapy.MkTreader.get` function doc. 
        The default is `None`
    `param` : `dict`, optional;
        Additional parameters required by mkt data provider.
        For more details see the `azapy.MkTreader.get` function doc. 
        The default is `None`.
    `except_file` : `list`, optional;
        List of symbols to be omitted from the update. The default is [].
    `verbose` : Boolean, optional;
    
        - `True` will print a progress report,
        - `False` suppress any printing to the terminal.
        
        The default is `True`.

    Returns
    -------
    Error code (int):
        - 200 : successful, everything updated
        - 201 : some (or all) were not completly updated 
        - 101 : the `mktdir` does not exists
        - 102 : unsupported mkt data sources
            
    Note that files with unsupported extensions 
    (see `azapy.MkTreader.get` function)
    are silently omitted from the update.
    '''
    
    if not os.path.isdir(mktdir):
        warnings.warn(f"Warning: doesn\'t exist dir: {mktdir}")
        return 101
    if source is None:
        source = 'yahoo'
    elif not source in ['yahoo', 
                        'eodhistoricaldata', 'eodhistoricaldata_yahoo',
                        'alphavantage', 'alphavantage_yahoo',
                        'marketstack']:
        warnings.warn(f"Unsupported mkt data source {source}")
        return 102
    if except_file is None:
        except_file = []
    
    sce = {}
    for file in os.listdir(mktdir):
        if verbose:
            print(file)
        fname, fext = os.path.splitext(file)
        if fext[1:] not in {'csv', 'json', 'feather'}:
            if verbose:
                print(f"file {file} omitted from update")
            continue
        if fname in except_file:
            if verbose:
                print(f"file {file} excepted from update")
            continue
        ld = {}
        ld['source'] = source
        ld['file_format'] = fext[1:]
        sce[fname] = ld

    mkt = MkTreader()
    mkt.get(source=sce, file_dir=mktdir, api_key=api_key, 
            param=param, verbose=verbose)
    
    elo = mkt.get_error_log()
    if verbose:
        sta = mkt.get_request_status()
        print("\nThe following symbols were updated:")
        for symb in sta.columns:
            print(f"symb {symb} : errors {sta.loc['error', symb]}")
        if len(elo) != 0:
            print(f"errors: {elo}")
          
    if len(elo) == 0: 
        return 200
    return 201


"""
Wrappers for various solvers.
"""

import numpy as np
import scipy.sparse as sps
import scipy.optimize as sopt
import scipy.linalg as la
import warnings
import cvxopt as cx
import ecos

_tol_cholesky = 1.e-10

def _lp_scipy(c, G, h, A, b, method):
    # compute - suppress warning {'sparse': True}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        res = sopt.linprog(c=c, A_ub=G, b_ub=h, 
                           A_eq=A, b_eq=b, method=method, 
                           options={'sparse': True})
        
    # gather the results
    rout = {}
    rout['status'] = res.status
    rout['infostring'] = res.message
    rout['pcost'] = np.float64(res.fun)
    rout['x'] = res.x
    
    return rout

def _lp_cvxopt(c, G, h, A, b, method):
    c = cx.matrix(c)
    G = cx.spmatrix(G.data, G.row, G.col, size=G.shape)
    h = cx.matrix(h)
    if A is not None:
        A = cx.spmatrix(A.data, A.row, A.col, size=A.shape)
        b = cx.matrix(b)
    
    if method == 'cvxopt': 
        solver = None
    else: 
        solver = method
    
    res = cx.solvers.lp(c, G, h, A, b, solver=solver, 
                            options={'show_progress': False,
                                     'glpk':{'msg_lev':'GLP_MSG_OFF'}})
    
    # gather the results
    rout = {}
    rout['status'] = 0 if 'optimal' in res['status'] else 2
    rout['infostring'] = res['status']
    rout['pcost'] = np.float64(res['primal objective'])
    rout['x'] = res['x']
    
    return rout

def _lp_ecos(c, G, h, A, b, method):
    c = np.array(c)
    G = G.tocsc()
    h = np.array(h)
    if A is not None:
        A = A.tocsc()
        b = np.array(b)
    dims = {'l': G.shape[0], 'q': [], 'e': 0}
    
    kwargs = {'verbose': False}
    res = ecos.solve(c, G, h, dims, A, b, **kwargs)
    
    # gather the results
    rout = {}
    rout['status'] = res['info']['exitFlag']
    # accept 10 Close to optimal as optimal
    if rout['status'] == 10: rout['status'] = 0
    rout['infostring'] = res['info']['infostring']
    rout['pcost'] = np.float64(res['info']['pcost'])
    rout['x'] = res['x']
    
    return rout
    
def _lp_solver(method, c_data, G, h_data, A=None, b_data=None):
    if method == 'ecos':
        return _lp_ecos(c_data, G, h_data, A, b_data, method)
    elif method in ['glpk', 'conelp', 'cvxopt']:
        return _lp_cvxopt(c_data, G, h_data, A, b_data, method)
    else:
        return _lp_scipy(c_data, G, h_data, A, b_data, method)

def _socp_cvxopt(c, G, h, dims, A, b):
    c = cx.matrix(c)
    G = cx.spmatrix(G.data, G.row, G.col, size=G.shape)
    h = cx.matrix(h)
    dims['s'] = []
    if A is not None:
        A = cx.spmatrix(A.data, A.row, A.col, size=A.shape)
        b = cx.matrix(b)
        
    res = cx.solvers.conelp(c, G, h, dims, A, b, 
                            options={'show_progress': False})
    
    # gather the results
    rout = {}
    rout['status'] = 0 if 'optimal' in res['status'] else 2
    rout['infostring'] = res['status']
    rout['pcost'] = np.float64(res['primal objective'])
    rout['x'] = res['x']
    
    return rout

def _socp_ecos(c, G, h, dims, A, b):
    c = np.array(c)
    G = G.tocsc()
    h = np.array(h)
    dims['e'] = 0
    if A is not None:
        A = A.tocsc()
        b = np.array(b)
    
    kwargs = {'verbose': False, 'max_iters': 1000}
    res = ecos.solve(c, G, h, dims, A, b, **kwargs)
    
    # gather the results
    rout = {}
    rout['status'] = res['info']['exitFlag']
    # accept 10 Close to optimal as optimal
    if rout['status'] == 10: rout['status'] = 0
    rout['infostring'] = res['info']['infostring']
    rout['pcost'] = np.float(res['info']['pcost'])
    rout['x'] = res['x']
    
    return rout

def _socp_solver(method, c_data, G, h_data,dims, A=None, b_data=None):
    if method == 'ecos':
        return _socp_ecos(c_data, G, h_data, dims, A, b_data)
    else:
        return _socp_cvxopt(c_data, G, h_data, dims, A, b_data)

def _qp_cvxopt(P, q, G, h, A, b):
    P = cx.matrix(P)
    q = cx.matrix(q)
    G = cx.spmatrix(G.data, G.row, G.col, size=G.shape)
    h = cx.matrix(h)
    if A is not None:
        A = cx.spmatrix(A.data, A.row, A.col, size=A.shape)
        b = cx.matrix(b)
    # else:
    #     AA = None
    #     bb = None
    
    res = cx.solvers.qp(P, q, G, h, A, b, 
                        options={'show_progress': False})
    
    # gather the results
    rout = {}
    rout['status'] = 0 if 'optimal' in res['status'] else 2
    rout['infostring'] = res['status']
    rout['pcost'] = np.float64(res['primal objective'])
    rout['x'] = res['x']
    
    return rout

def _qp_ecos(P, q, G, h, A, b):
    cc = np.concatenate((q, [1.]))
    
    gr, gc = G.shape
    pr, pc = P.shape
    
    gg = sps.block_diag((G, [-1.]))
    
    if any(np.diag(P) < _tol_cholesky):
        pp = sps.block_diag((-la.sqrtm(P), [1.]))
    else:
        pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), [1.]))
        
    pp = sps.block_diag((-la.sqrtm(P), [1.]))
    GG = sps.vstack([gg, pp], format='csc')
  
    dims = {'l': gr, 'q': [pr + 2], 'e': 0}
    
    hh = np.concatenate((h, [1.] + [0.] * (pr + 1)))

    if A is not None:
        A = sps.hstack((A, np.zeros((A.shape[0], 1))), format='csc')
        b = np.array(b)

    kwargs = {'verbose': False}
    res = ecos.solve(cc, GG, hh, dims, A, b, **kwargs)
    
    # gather the results
    rout = {}
    rout['status'] = res['info']['exitFlag']
    # accept 10 Close to optimal as optimal
    if rout['status'] == 10: rout['status'] = 0
    rout['infostring'] = res['info']['infostring']
    rout['pcost'] = np.float64(res['info']['pcost'] + 0.5)
    rout['x'] = res['x'][:-1]
    
    return rout
    
def _qp_solver(method, P, q_data, G, h_data, A=None, b=None):
    if method == 'ecos':
        return _qp_ecos(P, q_data, G, h_data, A, b)
    else:
        return _qp_cvxopt(P, q_data, G, h_data, A, b)
    
    
def _exp_cone_ecos(c, G, h, dims, A, b):
    c = np.array(c)
    G = G.tocsc()
    h = np.array(h)

    if A is not None:
        A = A.tocsc()
        b = np.array(b)
    
    kwargs = {'verbose': False, 'max_iters': 1000}
    res = ecos.solve(c, G, h, dims, A, b, **kwargs)
    
    # gather the results
    rout = {}
    rout['status'] = res['info']['exitFlag']
    # accept 10 Close to optimal as optimal
    if rout['status'] == 10: rout['status'] = 0
    rout['infostring'] = res['info']
    rout['pcost'] = np.float64(res['info']['pcost'])
    rout['x'] = res['x']
 
    return rout


def _exp_cone_solver(method, c, G, h, dims, A=None, b=None):
    # only ecos so far - don't check for method
    return _exp_cone_ecos(c, G, h, dims, A, b)
    


from .Port_InvVol import Port_InvVol
from .KellyEngine import KellyEngine

class Port_Kelly(Port_InvVol):
    """
    Backtesting Kelly optimal portfolio, periodically rebalanced.
    
    Methods:
        * set_model
        * get_port
        * get_nshares
        * get_weights
        * get_account
        * get_mktdata
        * port_view
        * port_view_all
        * port_drawdown
        * port_perf
        * port_annual_returns
        * port_monthly_returns
        * port_period_returns
    """
    def set_model(self, rtype='ExpCone', hlength=1.25, method='ecos'):
        """
        Sets model parameters and evaluates portfolio time-series.

        Parameters
        ----------
        `rtype` : `str`, optional;
            Type of optimization. It could take the values:\n
                `'ExpCone'` - Exponential cone constraint programming solution
                for full Kelly problem. \n
                `'Full'` - Non-linear solver for full Kelly problem. \n
                `'Order2'` - Second order Tayler approximation of Kelly problem. \n
            The default is `'ExpCone'`.
        `hlength` : `float`, optional;
            The length in year of the historical calibration period relative 
            to `'Dfix'`. A fractional number will be rounded to an integer number 
            of months. The default is `1.25` years. 
        `method` : `str`, optional;
            The QP solver class. It is relevant only if `rtype='Order2'`.
            It takes 2 values: `'ecos'` or `'cvxopt'`.
            The default is `'ecos'`.

        Returns
        -------
        `pandas.DataFrame`;
            The portfolio time-series in the format "date", "pcolname".
        """
        self.rtype= rtype
        self.method = method
        
        return super().set_model(hlength)
 
            
    def _ww_calc(self, data):
        return KellyEngine().getWeights(rrate=data, 
                                        rtype=self.rtype, 
                                        method=self.method)



import numpy as np
import scipy.linalg as la
import scipy.sparse as sps
import warnings
import time

from ._RiskAnalyzer import _RiskAnalyzer
from ._solvers import _socp_solver, _qp_solver, _tol_cholesky


class MVAnalyzer(_RiskAnalyzer):
    """
    MV (mean-variance) - Variance based optimal portfolio strategies.
    
    Methods:
        * getWeights
        * getRisk
        * getPositions
        * getRiskComp
        * getDiversification
        * viewForntiers
        * set_rrate
        * set_mktdata
        * set_rtype
        * set_random_seed
    """
    def __init__(self, mktdata=None, colname='adjusted', freq='Q', 
                 hlength=3.25, calendar=None, rtype='Sharpe', method = 'ecos',
                 name='MV'):
        """
        Constructor

        Parameters
        ----------
       `mktdata` : `pandas.DataFrame`, optional;
           Historic daily market data for portfolio components in the format
           returned by `azapy.mktData` function. The default is `None`.
       `colname` : `str`, optional;
           Name of the price column from mktdata used in the weights 
           calibration. The default is `'adjusted'`.
       `freq` : `str`, optional;
           Rate of return horizon. It could be 
           `'Q'` for quarter or `'M'` for month. The default is `'Q'`.
       `hlength` : `float`, optional;
           History length in number of years used for calibration. A 
           fractional number will be rounded to an integer number of months.
           The default is `3.25` years.
       `calendar` : `numpy.busdaycalendar`, optional;
           Business days calendar. If is it `None` then the calendar will 
           be set to NYSE business calendar. 
           The default is `None`.
       `rtype` : `str`, optional;
           Optimization type. Possible values: \n
               `'Risk'` : optimal-risk portfolio for targeted expected rate of 
               return.\n
               `'Sharpe'` : Sharpe-optimal portfolio - maximization solution.\n
               `'Sharpe2'` : Sharpe-optimal portfolio - minimization solution.\n
               `'MinRisk'` : minimum risk portfolio.\n
               `'RiskAverse'` : optimal-risk portfolio for a fixed 
               risk-aversion factor.\n
               `'InvNrisk'` : optimal-risk portfolio with the same risk value 
               as a benchmark portfolio (e.g., same as equal weighted 
               portfolio).\n
               `'Diverse'` : optimal-diversified portfolio for targeted
               expected rate of return (maximum of inverse 1-D).\n
               `'Diverse2'` : optimal-diversified portfolio for targeted
               expected rate of return (minmum of 1-D).\n
               `'MaxDiverse'` : maximum diversified portfolio.\n
               `'InvNdiverse'` : optimal-diversified portfolio with the same
               diversification factor as a benchmark portfolio 
               (e.g., same as equal weighted portfolio).\n
               `'InvNdrr'` : optima- diversified portfolio with the same 
               expected rate of return as a benchmark portfolio
               (e.g., same as equal weighted portfolio).\n
           The defauls is `'Sharpe'`.
        `method` : `str`, optional;
            Quadratic programming numerical method. Could be `'ecos'` or
            'cvxopt'. The default is `'ecos'`.
        `name` : `str`, optional;
            Object name. The default is `'MV'`.
            
        Returns
        -------
        The object.
        """
        super().__init__(mktdata, colname, freq, hlength, calendar, 
                         rtype, name)
        
        self._set_method(method)
        
        
    def _set_method(self, method):
        self._set_qp_method(method)
    
        
    def _risk_calc(self, prate, alpha):
        var = np.var(prate)
        
        # status, volatility, variance,
        return 0, np.sqrt(var), var
    
    
    def _risk_min(self, d=1):
        # Order of variables
        # w <- [0:nn]
        # in total dim=nn
        
        # build P
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build q
        q_data = [0.] * nn
        
        # build G
        icol = list(range(nn)) + list(range(nn))
        irow = [0] * nn + list(range(1, nn + 1))
        data = list(-self.muk * d) + [-1.] * nn
        
        G = sps.coo_matrix((data, (irow, icol)), shape=(nn + 1, nn))
        
        # build h
        h_data = [-self.mu * d] + [0.] * nn
        
        # build A
        A = sps.coo_matrix([1.] * nn)
        
        # build
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _qp_solver(self.method, P, q_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
 
        # optimal weights
        self.ww = np.array(res['x'])
        self.ww.shape = nn
        # variance
        self.risk = 2 * res['pcost']
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(np.abs(self.risk))])
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        
        return self.ww
    
    
    def _sharpe_max(self):
        # Computes the mazimization of Sharpe
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # in total dim = nn + 1
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c 
        c_data = list(-self.muk) + [self.mu] 

        # biuld G
        # ww > 0 and t > 0
        dd = sps.diags([-1] * (nn + 1), format='coo')
        
        # cone
        xx = sps.coo_matrix(([-1.], ([0], [nn])), shape=(1, nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = sps.block_diag((-la.sqrtm(P), [-1.]))
        else:
            pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), [-1.]))
            
        G = sps.vstack([dd, xx, pp])
        
        # biuld dims
        dims = {'l': nn + 1, 'q': [nn + 2]}
        
        # build h
        h_data = [0.] * (nn + 1) + [0.25] + [0.] * nn + [-0.25]
        
        # build A
        A = sps.coo_matrix([1.] * nn + [-1.])
            
        # build b
        b_data = [0.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
 
        t = res['x'][-1]
        # sharpe
        self.sharpe = -res['pcost']
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # rate of return
        self.RR = self.mu -  res['pcost'] / t
        #self.RR = np.dot(self.ww, self.muk)
        # variance
        self.risk = 1. / t
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        
        return self.ww
    
    
    def _sharpe_inv_min(self):
        # Computes the minimum of inverse Sharpe
        # Order of variables
        # w <- [0:nn]
        # u <- nn
        # t <- nn + 1
        # in total dim = nn + 2
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = [0.] * nn + [1., 0.]
        sq2 = -np.sqrt(0.5)

        # build G
        # ww > 0 and u > 0 and t > 0
        dd = sps.diags([-1] * (nn + 2), format='coo')
        
        # cone
        xx = sps.coo_matrix(([sq2, sq2], ([0, 0], [nn, nn + 1])), 
                            shape=(1, nn + 2))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = sps.block_diag((-la.sqrtm(P), np.diag([sq2, sq2])))
        else:
            pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), 
                                  np.diag([sq2, sq2])))
            
        G = sps.vstack([dd, xx, pp])
        
        # build h
        h_data = [0.] * (2 * nn + 5)
        
        # def dims
        dims = {'l': nn + 2, 'q': [nn + 3]}
        
        # build A
        A = sps.coo_matrix(
            [[1.] * nn + [0., -1.], list(self.muk) + [0., -self.mu]])
 
        # build b
        b_data = [0., 1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
  
        t = res['x'][-1]
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # MV-Sharpe
        self.sharpe = 1. / res['pcost']
        # variance
        self.risk = res['pcost'] / t
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        # rate of return
        self.RR = 1. / t + self.mu
        #self.RR = np.dot(self.ww, self.muk)
        
        return self.ww   
    
    
    def _rr_max(self):
        # Computes the maximization of returns (for fixed volatility)
        # Order of variables
        # w <- [0:nn]
        # in total dim = nn 
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk)
        
        # build G
        # ww >= 0 + 1 cone
        dd = sps.diags([-1.] * nn, shape=(nn + 1, nn), format='coo')
        
        if any(np.diag(P) < _tol_cholesky):
            G = sps.vstack([dd, -la.sqrtm(P)])
        else:
            G = sps.vstack([dd, -la.cholesky(P, overwrite_a=True)])
        
        # build h
        h_data = [0.] * nn + [np.sqrt(self.risk)] + [0.] * nn
        
        # def dims
        dims = {'l': nn, 'q': [nn + 1]}
        
        # build A_eq
        A = sps.coo_matrix([1.] * nn) 
        
        # build b_eq
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
        
        # optimal weights
        self.ww = np.array(res['x'][:])
        self.ww.shape = nn
        # rate of return
        self.RR = -res['pcost']
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        
        return self.ww   
    
    
    def _risk_averse(self):
        # Order of variables
        # w <- [0:nn]
        # in total dim=nn
        
        # build P
        P = self.rrate.cov().to_numpy() * (2. * self.Lambda)
        nn = P.shape[0]
 
        # build q
        q_data = list(-self.muk)
        
        # build G
        G = sps.diags([-1.] * nn, format='coo')
        
        # build h
        h_data = [0.] * nn
        
        # build A
        A = sps.coo_matrix([1.] * nn)
        
        # build
        b_data = [1.]
        
        # calc
        toc = time.perf_counter()
        res = _qp_solver(self.method, P, q_data, G, h_data, A, b_data)
        self.time_level2 = time.perf_counter() - toc
        
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
        
        # optimal weights
        self.ww = np.array(res['x'])
        self.ww.shape = nn
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        # variance
        self.risk = (res['pcost'] + self.RR) / self.Lambda
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        
        return self.ww
    
    
    def _risk_diversification(self, d=1):
        # Computes the minimum of inverse Sharpe
        # Order of variables
        # w <- [0:nn]
        # u <- nn
        # t <- nn + 1
        # in total dim = nn + 2
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = [0.] * nn + [1., 0.]
        sq2 = -np.sqrt(0.5)

        # build G
        yy = sps.coo_matrix(list(-self.muk * d) + [0., self.mu * d])
        
        # ww > 0 and u > 0 and t > 0
        dd = sps.diags([-1] * (nn + 2))
        
        # cone
        xx = sps.coo_matrix(([sq2, sq2], ([0, 0], [nn, nn + 1])), 
                            shape=(1, nn + 2))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = sps.block_diag((-la.sqrtm(P), np.diag([sq2, sq2])))
        else:
            pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), 
                                  np.diag([sq2, sq2])))
            
        G = sps.vstack([yy, dd, xx, pp])
        
        # build h
        h_data = [0.] * (2 * nn + 6)
        
        # def dims
        dims = {'l': nn + 3, 'q': [nn + 3]}
        
        # build A
        A = sps.coo_matrix(
            [[1.] * nn + [0., -1.], list(self.risk_comp) + [0., 0.]])
 
        # build b
        b_data = [0., 1.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
  
        t = res['x'][-1]
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # MV-Diverse
        self.diverse = 1. - res['pcost']
        # variance
        self.risk = res['pcost'] / t
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        
        return self.ww   
    
    
    def _risk_inv_diversification(self, d=1):
        # Computes the minimum of inverse Sharpe
        # Order of variables
        # w <- [0:nn]
        # u <- nn
        # t <- nn + 1
        # in total dim = nn + 2
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.risk_comp) + [0.] 
        
        # biuld G
        icol = list(range(nn + 1)) + list(range(nn + 1))
        irow = [0] * (nn + 1) + list(range(1, nn + 2)) 
        data = list(-self.muk * d) + [self.mu * d] + [-1.] * (nn + 1)
        dd = sps.coo_matrix((data, (irow, icol)), shape=(nn + 2, nn + 1))
        
        # cone
        xx = sps.coo_matrix(([-1.], ([0], [nn])), shape=(1, nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = sps.block_diag((-la.sqrtm(P), [-1.]))
        else:
            pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), [-1.]))
            
        G = sps.vstack([dd, xx, pp])
        
        # biuld dims
        dims = {'l': nn + 2, 'q': [nn + 2]}
        
        # build h
        h_data = [0.] * (nn + 2) + [0.25] + [0.] * nn + [-0.25]
        
        # build A
        A = sps.coo_matrix([1.] * nn + [-1.])
            
        # build b
        b_data = [0.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
 
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
  
        t = res['x'][-1]
        # optimal weights
        self.ww = np.array(res['x'][:nn]) / t
        self.ww.shape = nn
        # MV-Diverse
        self.diverse = 1. + 1. / res['pcost']
        # variance
        self.risk = 1. / t
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        # rate of return
        self.RR = np.dot(self.ww, self.muk)
        
        return self.ww   


    def _rr_max_diversification(self):
        # Computes the maximization of returns (for fixed volatility)
        # Order of variables
        # w <- [0:nn]
        # t <- nn
        # in total dim = nn + 1
        P = self.rrate.cov().to_numpy()
        nn = P.shape[0]
        
        # build c
        c_data = list(-self.muk) +[0.]
        
        # build G
        # ww >= 0 snd t > 0
        dd = sps.diags([-1.] * (nn + 1), shape=(nn + 1, nn  + 1), format='coo')
        
        # cone
        xx = sps.coo_matrix(([-1.], ([0], [nn])), shape=(1, nn + 1))
        
        if any(np.diag(P) < _tol_cholesky):
            pp = sps.block_diag((-la.sqrtm(P), [-1.]), format='coo')
        else:
            pp = sps.block_diag((-la.cholesky(P, overwrite_a=True), [-1.]), format='coo')
            
        G = sps.vstack([dd, xx, pp])
        
        # biuld dims
        dims = {'l': nn + 1, 'q': [nn + 2]}
        
        # build h
        h_data = [0.] * (nn + 1) + [0.25] + [0.] * nn + [-0.25]
        
        # build A_eq
        A_icol = list(range(nn)) + list(range(nn + 1))
        A_irow = [0] * nn + [1] * (nn + 1)
        A_data = [1.] * nn + list((1. - self.diverse) * self.risk_comp) + [-1.]
        A = sps.coo_matrix((A_data, (A_irow, A_icol)), shape=(2, nn + 1)) 
        
        # build b_eq
        b_data = [1., 0.]
        
        # calc
        toc = time.perf_counter()
        res = _socp_solver(self.method, c_data, G, h_data, dims, A, b_data)
        self.time_level2 = time.perf_counter() - toc
    
        self.status = res['status']
        if self.status != 0:
            warnings.warn(f"Warning {self.name} on {self.rrate.index[-1]} :: "
                          f"status {res['status']} :: {res['infostring']}")
            return np.array([np.nan] * nn)
        
        # optimal weights
        self.ww = np.array(res['x'][:-1])
        self.ww.shape = nn
        # rate of return
        self.RR = -res['pcost']
        # risk
        self.risk = np.dot(np.dot(self.ww, P), self.ww)
        # variance
        self.primary_risk_comp = np.array([self.risk])
        # volatility
        self.secondary_risk_comp = np.array([np.sqrt(self.risk)])
        # diversification
        self.diverse = 1. - self.risk / np.dot(self.ww, self.risk_comp)
        
        return self.ww   